#!/usr/bin/python
# iinterface.py
# 2004 by Erik Andresen / Defiant <erik@defiant.homedns.org>
#
# Most serial part is based on the code from Grant B. Edwards  <grante@visi.com>, 
# Posix serial port class found on ftp://ftp.visi.com/users/grante/python/PosixSerial.py
#
# thanks also to #python.de in the ircNET.

import termios
import os
import struct
import time
from threading import Thread


# construct dictionaries for baud rate lookups
baudEnumToInt = {}
baudIntToEnum = {}
for rate in (0,50,75,110,134,150,200,300,600,1200,1800,2400,4800,9600,19200,38400,57600,115200,230400,460800,500000,576000,921600,1000000,1152000,1500000,2000000,2500000,3000000,3500000,4000000):
    try:
        i = eval('termios.B'+str(rate))
        baudEnumToInt[i]=rate
        baudIntToEnum[rate] = i
    except:
        pass

class IInterface(Thread):
	# Init: open the serial console and start the thread
	def __init__(self, Device="/dev/ttyS0", numIfaces = 1, StartMe = 1):
		self.myInterface = os.open(Device, os.O_RDWR | os.O_NOCTTY)
		
		iflag, oflag, cflag, lflag, ispeed, ospeed, cc = termios.tcgetattr(self.myInterface)

		ispeed = ospeed = baudIntToEnum[9600]
		cflag |=  (termios.CS8 | termios.CRTSCTS | termios.CLOCAL | termios.CREAD)
		oflag = termios.OPOST
		lflag = termios.OPOST
		cc[termios.VTIME] = 1
		cc[termios.VMIN] = 0
		
		termios.tcsetattr(self.myInterface, termios.TCSANOW, [iflag, oflag, cflag, lflag, ispeed, ospeed, cc])
	
		# small test
		#self.write(0xC1)
		#self.write(0x00)
		#if not self.read():
		#	print "Warning: No Interface!"
		
		Thread.__init__(self)
		self.setDaemon(1)
		if (StartMe == 1):
			self.start()
		
		self.myExit = 0
		self.numIfaces = numIfaces
		self.MotorStatus1 = 0x0
		self.MotorStatus2 = 0x0
		self.Digitals1 = 0x0
		self.Digitals2 = 0x0
		self.AnaX = 0x0
		self.AnaY = 0x0
		self.MotorSpeed = {}
		self.curSpeed = {}
		self.INTERFACE_QUERY_TIME = 0.01
		
		time.sleep(1)
	
	# Just cycle ReadX & Y to make sure the digitals, X, Y 
	# and that the engine states are constantly set
	def run(self):
		i = 0
		while(self.myExit == 0):
			if i == 10:
				self.ReadX()
			elif i == 20:
				self.ReadY()
			else:
				self.ReadDigitals()
			
			i+=1
			self.CycleMotorSpeed()
			time.sleep(self.INTERFACE_QUERY_TIME)

	def exit(self):
		# TODO: Some more stuff for cleanup?
		os.close(self.myInterface)
		self.myExit = 1

	def write(self, Num):
		struct.pack('B', Num)
		os.write(self.myInterface, struct.pack('B', Num))
		# alternative is os.write(self.myInterface, chr(Num))

	def read(self, Num=1):
		StrRead = None
		StrRead = os.read(self.myInterface, Num)
		# alternative is ord(os.read(self.myInterface, 1))
		if StrRead == "":
		#	print "Warning: No connection to the Interface"
			return None
		if Num == 2:
			try:
				#print 'foo: ', len(StrRead), ord(StrRead)
				list = struct.unpack('BB', StrRead)
				muh = list[0]
				kuh = list[1]
				print kuh, muh, kuh<<4
				kuh<<=4
				print 'final', kuh + muh
			except:
				#print 'error'
				return
			return struct.unpack('H', StrRead)[0]
		else:
			return struct.unpack('B', StrRead)[0]
	
	# Only read Digitals, no X, no Y - not used
	def ReadDigitals(self):
		if self.numIfaces == 2:
			askNum = 0xC2
		else:
			askNum = 0xC1
		
		self.write(askNum)
		self.write(self.MotorStatus1)
		if self.numIfaces == 2:
			self.write(self.MotorStatus2)
		self.Digitals1 = self.read()
		if self.numIfaces == 2:
			self.Digitals2 = self.read()
	
	# Reads AnaX, Digital Input and reset the engine state
	def ReadX(self):
		if self.numIfaces == 2:
			askNum = 0xC6
		else:
			askNum = 0xC5
		
		self.write(askNum)
		self.write(self.MotorStatus1)
		if self.numIfaces == 2:
			self.write(self.MotorStatus2)
		self.Digitals1 = self.read()
		if self.numIfaces == 2:
			self.Digitals2 = self.read()
		self.AnaX = self.read(2)

	# Reads AnaY, Digital Input and reset the engine state
	def ReadY(self):
		if self.numIfaces == 2:
			askNum = 0xCA
		else:
			askNum = 0xC9
		
		self.write(askNum)
		self.write(self.MotorStatus1)
		if self.numIfaces == 2:
			self.write(self.MotorStatus2)
		self.Digitals1 = self.read()
		if self.numIfaces == 2:
			self.Digitals2 = self.read()
		self.AnaY = self.read(2)

	# Set Motor state. Direction l,r,s for left, right, stop
	def SetMotor(self, Motor, direction, overrideSpeed=1):
		stop = 0x03

		if overrideSpeed == 1:
			self.MotorSpeed[Motor] = 0
			self.curSpeed[Motor] = 0
		
		if Motor > 4:
			add = 1
			Motor-=4
			MotorStatus = self.MotorStatus2
		else:
			add = 0
			MotorStatus = self.MotorStatus1


		switch = {
			'l': 0x1,
			'r': 0x2,
			's': 0x0
		}
		direction = switch[direction]

		# Reset engine to stop:
		stop<<=(Motor-1)*2
		# choose right engine and use OR to target it
		MotorStatus = stop | MotorStatus
		# now invert 11 to 00 with a simple XOR
		MotorStatus = stop ^ MotorStatus
		
		# choose right engine
		direction<<=(Motor-1)*2
		
		MotorStatus = direction | MotorStatus
		
		if add == 1:
			self.MotorStatus2 = MotorStatus
		else:
			self.MotorStatus1 = MotorStatus

	# Returns the current engine State
	def GetMotor(self, Motor):
	
		if Motor > 4:
			Motor-=4
			MotorStatus = self.MotorStatus2
		else:
			MotorStatus = self.MotorStatus1
	
		# Target the bytes, todo it we have to get them to the end...
		# Note: maybe it is easier if we just devide with two and take the rest.
		state = MotorStatus<<(4-Motor)*2
		# and now back to the beginning
		state>>=6
		
		switch = {
			0x1: 'l',
			0x2: 'r',
			0x0: 's'
		}
		return switch[state]

	# is looking at the dezimal value and converting to a value 1 oder 0
	def Digital(self, Input):
		c=0x1
		if Input > 8:
			Input-=8
			Digitals = self.Digitals2
		else:
			Digitals = self.Digitals1
		if not Digitals:
			return
		c<<=(Input-1)
		c&=Digitals
		c>>=(Input-1)

		return c
		
	def GetX(self):
		return self.AnaX
	def GetY(self):
		return self.AnaY

	def SetMotorSpeed(self, Motor, Speed):
		if Speed == 0 and self.MotorSpeed.has_key(Motor):
			del self.MotorSpeed[Motor]
			return
		self.MotorSpeed[Motor] = Speed
		self.curSpeed[Motor] = 0
	
	def GetMotorSpeed(self, Motor):
		if not self.MotorSpeed.has_key(Motor):
			return 0
		return self.MotorSpeed[Motor]

	def CycleMotorSpeed(self):
		for Motor in range(self.numIfaces * 4):
			Motor+=1
			if not self.MotorSpeed.has_key(Motor):
				continue
			
			if self.MotorSpeed[Motor] > 0:
				if self.curSpeed[Motor] <= self.MotorSpeed[Motor]:
					self.SetMotor(Motor, 'l', 0)
				else:
					self.SetMotor(Motor, 's', 0)
				if  self.curSpeed[Motor] >= 10:
					self.curSpeed[Motor] = 1
				else:
					self.curSpeed[Motor]+=1
			elif self.MotorSpeed[Motor] < 0:
				if self.curSpeed[Motor] >= self.MotorSpeed[Motor]:
					self.SetMotor(Motor, 'r', 0)
				else:
					self.SetMotor(Motor, 's', 0)
				if  self.curSpeed[Motor] <= -10:
					self.curSpeed[Motor] = -1
				else:
					self.curSpeed[Motor]-=1

